import { updatePlanetResources } from "../../db/planets";
import { getAllResources } from "../../db/resources";
import { Planet } from "./PlanetManager";
import ResourceManager from "./abstract/ResourceManager";

export default class PlanetResourceManager extends ResourceManager {
    planet: Planet;

    constructor(planet: Planet) {
        super();
        this.planet = planet;
    }

    async init(resourceData: { id: string, amount: number, lastUpdated: Date }[]) {
        const resources = await getAllResources();

        this.resourcesDB = resources;

        if(resourceData.length === 0) {
            resourceData = [
                {
                    id: "coal",
                    amount: 11,
                    lastUpdated: new Date()
                },
                {
                    id: "iron",
                    amount: 22,
                    lastUpdated: new Date()
                },
                {
                    id: "gold",
                    amount: 33,
                    lastUpdated: new Date()
                },
                {
                    id: "water",
                    amount: 44,
                    lastUpdated: new Date()
                },
                {
                    id: "sulfuric-acid",
                    amount: 55,
                    lastUpdated: new Date()
                },
                {
                    id: "liquid-nitrogen",
                    amount: 66,
                    lastUpdated: new Date()
                },
                {
                    id: "hydrogen",
                    amount: 77,
                    lastUpdated: new Date()
                },
                {
                    id: "oxygen",
                    amount: 88,
                    lastUpdated: new Date()
                },
                {
                    id: "helium-3",
                    amount: 99,
                    lastUpdated: new Date()
                }
            ];

            await updatePlanetResources(this.planet._id, this.resources);
        }
    
        resourceData.forEach(resource => {
            const resFromDB = resources.find(res => res.id === resource.id);
            if(resFromDB) this.resources.push({
                id: resource.id,
                amount: resource.amount,
                lastUpdated: resource.lastUpdated,
                data: resFromDB
            })
        });

        return this;
    }

    getStorageCapacities() {
        const mapStorageToResource: { [key: string]: string } = {
            "coal-storage": "coal",
            "iron-storage": "iron",
            "gold-storage": "gold",
            "water-tank": "water",
            "acid-tank": "sulfuric-acid",
            "nitrogen-tank": "liquid-nitrogen",
            "hydrogen-tank": "hydrogen",
            "oxygen-tank": "oxygen",
            "helium3-tank": "helium-3",
            "uranium-container": "uranium",
            "oil-tank": "oil",
            "diesel-tank": "diesel",
            "kerosene-tank": "kerosene",
            "deuterium-tank": "deuterium",
            "tritium-tank": "tritium",
        }

        const output: { id: string, capacity: number }[] = [];
    
        for(const building of this.planet.buildings.buildings) {
            if(building.data.category === 'storage') {
                const resource = mapStorageToResource[building.data.id];
                if(resource) {
                    output.push({ id: resource, capacity: building.level * 10_000 });
                }
            }
        }

        return output;
    }

    getPerHourMiningRate() {
        const mapMinerToResource: { [key: string]: string } = {
            "coal-mine": "coal",
            "iron-mine": "iron",
            "gold-mine": "gold"
        }

        const miningEfficiencyResearch = this.planet.system.data.ownedBy.research.getResearchById('mining-efficiency')?.level ?? 0;

        const output = [];
        const buildings = this.planet.buildings.buildings;

        for(const building of buildings) {
            if(Object.keys(mapMinerToResource).includes(building.data.id)) {
                const miningRate = Math.floor((20 * Math.pow(building.level, 2) + 40 * building.level + 15) * (1 + miningEfficiencyResearch * 0.05));
                
                output.push({
                    id: mapMinerToResource[building.data.id],
                    rate: miningRate
                });
            }
        }

        return output;
    }

    async sync() {
        await updatePlanetResources(this.planet._id, this.resources.map(res => {
            return {
                id: res.id,
                amount: res.amount,
                lastUpdated: res.lastUpdated,
            }
        }));
    }
}